@@ -17,14 +17,14 @@ use crate::data_types::UnsignedLike;
 /// The reader is consider is considered "aligned" if the current bit index
 /// is 0 or 8 (i.e. at the start or end of the current byte).
 #[derive(Clone)]
-pub struct BitReader {
-  bytes: Vec<u8>,
+pub struct BitReader<'a> {
+  bytes: &'a [u8],
   i: usize,
   j: usize,
   total_bits: usize,
 }
 
-impl Debug for BitReader {
+impl<'a> Debug for BitReader<'a> {
   fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
     let current_info = if self.i < self.bytes.len() {
       format!(
@@ -46,8 +46,8 @@ impl Debug for BitReader {
   }
 }
 
-impl From<Vec<u8>> for BitReader {
-  fn from(bytes: Vec<u8>) -> BitReader {
+impl<'a> From<&'a [u8]> for BitReader<'a> {
+  fn from(bytes: &'a [u8]) -> BitReader<'a> {
     let total_bits = 8 * bytes.len();
     BitReader {
       bytes,
@@ -58,11 +58,17 @@ impl From<Vec<u8>> for BitReader {
   }
 }
 
-impl BitReader {
+impl<'a, B: AsRef<[u8]>> From<&'a B> for BitReader<'a> {
+  fn from(bytes: &'a B) -> BitReader<'a> {
+    bytes.as_ref().into()
+  }
+}
+
+impl<'a> BitReader<'a> {
   /// Returns the reader's current byte index. Will return an error if the
   /// reader is at
   /// a misaligned position.
-  pub fn aligned_byte_ind(&self) -> QCompressResult<usize> {
+  pub fn aligned_byte_idx(&self) -> QCompressResult<usize> {
     if self.j == 0 {
       Ok(self.i)
     } else if self.j == 8 {
@@ -75,11 +81,15 @@ impl BitReader {
       )))
     }
   }
+  
+  pub(crate) fn bit_idx(&self) -> usize {
+    8 * self.i + self.j
+  }
 
   /// Returns the number of bits between the reader's current position and
   /// the end.
   pub fn bits_remaining(&self) -> usize {
-    self.total_bits - 8 * self.i - self.j
+    self.total_bits - self.bit_idx()
   }
 
   /// Returns the number of bytes in the reader.
@@ -88,7 +98,7 @@ impl BitReader {
   }
 
   /// Returns the reader's current (byte_idx, bit_idx) tuple.
-  pub fn inds(&self) -> (usize, usize) {
+  pub fn idxs(&self) -> (usize, usize) {
     (self.i, self.j)
   }
 
@@ -166,7 +176,7 @@ impl BitReader {
   }
 
   pub(crate) fn read_diff<Diff: UnsignedLike>(&mut self, n: usize) -> QCompressResult<Diff> {
-    if self.i * 8 + self.j + n > self.total_bits {
+    if self.bit_idx() + n > self.total_bits {
       return Err(QCompressError::insufficient_data(
         "read_diff(): reached end of data available to BitReader"
       ))
@@ -348,7 +358,7 @@ mod tests {
   fn test_bit_reader() -> QCompressResult<()>{
     // bits: 1001 1010  0110 1011  0010 1101
     let bytes = vec![0x9a, 0x6b, 0x2d];
-    let mut bit_reader = BitReader::from(bytes);
+    let mut bit_reader = BitReader::from(&bytes);
     assert_eq!(
       bit_reader.read_aligned_bytes(1)?,
       vec![0x9a],
@@ -377,18 +387,19 @@ mod tests {
 
   #[test]
   fn test_seek_rewind() {
-    let mut reader = BitReader::from(vec![0; 6]);
+    let bytes = vec![0; 6];
+    let mut reader = BitReader::from(&bytes);
     reader.seek(43);
 
     reader.rewind(2);
-    assert_eq!(reader.inds(), (5, 1));
+    assert_eq!(reader.idxs(), (5, 1));
     reader.rewind(2);
-    assert_eq!(reader.inds(), (4, 7));
+    assert_eq!(reader.idxs(), (4, 7));
     reader.rewind(7);
-    assert_eq!(reader.inds(), (4, 0));
+    assert_eq!(reader.idxs(), (4, 0));
     reader.rewind(8);
-    assert_eq!(reader.inds(), (3, 0));
+    assert_eq!(reader.idxs(), (3, 0));
     reader.rewind(17);
-    assert_eq!(reader.inds(), (0, 7));
+    assert_eq!(reader.idxs(), (0, 7));
   }
 }
