@@ -9,7 +9,7 @@ use crate::errors::{QCompressResult, QCompressError};
 /// 
 /// This is the part of chunk metadata that describes *how* the data was
 /// compressed - the Huffman codes used and what ranges they specify.
-#[derive(Clone, Debug)]
+#[derive(Clone, Debug, PartialEq)]
 pub enum PrefixMetadata<T: NumberLike> {
   /// `Simple` prefix metadata corresponds to the case when delta encoding is
   /// off (`delta_encoding_order` of 0).
@@ -42,7 +42,7 @@ pub enum PrefixMetadata<T: NumberLike> {
 /// be done easily - see the fast_seeking.rs example.
 /// One can also create a rough histogram (or a histogram of deltas, if
 /// delta encoding was used) by aggregating chunk metadata.
-#[derive(Clone, Debug)]
+#[derive(Clone, Debug, PartialEq)]
 pub struct ChunkMetadata<T> where T: NumberLike {
   /// The count of numbers in the chunk.
   pub n: usize,
@@ -138,8 +138,8 @@ impl<T> ChunkMetadata<T> where T: NumberLike {
   }
 
   pub fn write_to(&self, writer: &mut BitWriter, flags: &Flags) {
-    writer.write_usize(self.n, BITS_TO_ENCODE_N_ENTRIES as usize);
-    writer.write_usize(self.compressed_body_size, BITS_TO_ENCODE_COMPRESSED_BODY_SIZE as usize);
+    writer.write_usize(self.n, BITS_TO_ENCODE_N_ENTRIES);
+    writer.write_usize(self.compressed_body_size, BITS_TO_ENCODE_COMPRESSED_BODY_SIZE);
     match &self.prefix_metadata {
       PrefixMetadata::Simple { prefixes} => {
         write_prefixes(prefixes, writer, flags);
@@ -152,10 +152,13 @@ impl<T> ChunkMetadata<T> where T: NumberLike {
     writer.finish_byte();
   }
 
-  pub(crate) fn update_write_compressed_body_size(&self, writer: &mut BitWriter, initial_idx: usize) {
-    writer.assign_usize(
-      initial_idx + BITS_TO_ENCODE_N_ENTRIES as usize / 8,
-      BITS_TO_ENCODE_N_ENTRIES as usize % 8,
+  pub(crate) fn update_write_compressed_body_size(
+    &self,
+    writer: &mut BitWriter,
+    bit_idx: usize,
+  ) {
+    writer.overwrite_usize(
+      bit_idx + BITS_TO_ENCODE_N_ENTRIES,
       self.compressed_body_size,
       BITS_TO_ENCODE_COMPRESSED_BODY_SIZE,
     );
