@@ -1,12 +1,13 @@
 //! For crate-level documentation, see either
 //! <https://crates.io/crates/q_compress> or
 //! <https://github.com/mwlon/quantile-compression>.
+#[doc = include_str!("../README.md")]
 
 pub use bit_reader::BitReader;
 pub use bit_writer::BitWriter;
 pub use chunk_metadata::{ChunkMetadata, DecompressedChunk, PrefixMetadata};
 pub use compressor::{Compressor, CompressorConfig};
-pub use decompressor::{Decompressor, DecompressorConfig};
+pub use decompressor::{Decompressor, ChunkBodyDecompressor, DecompressorConfig};
 pub use flags::Flags;
 pub use prefix::Prefix;
 
@@ -18,6 +19,7 @@ mod bit_writer;
 mod bits;
 mod chunk_metadata;
 mod constants;
+mod compression_table;
 mod compressor;
 mod decompressor;
 mod delta_encoding;
@@ -25,6 +27,7 @@ mod flags;
 mod huffman_decoding;
 mod huffman_encoding;
 mod prefix;
+mod prefix_optimization;
 
 #[cfg(test)]
 mod tests;
